import java.io.*;
import java.lang.*;
import java.util.*;

/**
 * Class for reading datasets from files. Every dataset is defined by a base name and consists of two files,
 * <c>basename-knoten.csv</c> and <c>basename-kanten.csv</c>, respectively.
 *
 * Both files must be plain text files containing integer data seperated by semicolons. First line contains
 * fieldnames and is skipped on reading the file.
 *
 * @author Stefan Daschek
 */
public class NetData {

  /** HashMap holding knot-data. */
  HashMap knots;

  /** HashMap holding edge-data. */
  HashMap edges;

  
  /**
   * Read data from Files specified by baseFileName. Filenames are generated by appending "-knoten.csv" and
   * "-kanten.cvs" to baseFileName, respectively.
   *
   * @param  baseFileName Basename of files to read.
   * @throws FileNotFoundException  If any of the data files could not be found.
   * @throws NumberFormatException  If there is an error in any of the data files.
   * @throws IOException            If there is a problem reading any of the data files.
   */
  public NetData( String baseFileName )
    throws FileNotFoundException, NumberFormatException, NoSuchElementException, IOException
  {
    knots = new HashMap();
    edges = new HashMap();
    String line;
    
    // Read in knot data
    BufferedReader knotsReader = new BufferedReader( new FileReader( baseFileName + "-knoten.csv" ) );
    knotsReader.readLine();                                 // skip first line
    while( null != (line = knotsReader.readLine()) ) {
      int id, x, y, val1, val2;
      StringTokenizer csvTokenizer = new StringTokenizer( line, ";" );
      id   = Integer.parseInt( csvTokenizer.nextToken() );
      x    = Integer.parseInt( csvTokenizer.nextToken() );
      y    = Integer.parseInt( csvTokenizer.nextToken() );
      val1 = Integer.parseInt( csvTokenizer.nextToken() );
      val2 = Integer.parseInt( csvTokenizer.nextToken() );
      knots.put( new Integer(id), new Knot( id, x, y, val1, val2 ) );
    }
      
    // Read in edge data
    BufferedReader edgeReader = new BufferedReader( new FileReader( baseFileName + "-kanten.csv" ) );
    edgeReader.readLine();                                 // skip first line
    while( null != (line = edgeReader.readLine()) ) {
      int id, from, to, val;
      StringTokenizer csvTokenizer = new StringTokenizer( line, ";" );
      id   = Integer.parseInt( csvTokenizer.nextToken() );
      from = Integer.parseInt( csvTokenizer.nextToken() );
      to   = Integer.parseInt( csvTokenizer.nextToken() );
      val  = Integer.parseInt( csvTokenizer.nextToken() );
      edges.put( new Integer(id), new Edge( id, from, to, val ) );
    }
  }


  
  /**
   * Return HashMap holding knot data.
   *
   * @return HashMap holding knot data. 
   */
  public HashMap getKnots() {
    return knots;
  }


  
  /**
   * Return HashMap holding edge data.
   *
   * @return HashMap holding edge data. 
   */
  public HashMap getEdges() {
    return edges;
  }



  /**
   * Return Knot object for knot with specified id.
   *
   * @param id ID of the knot.
   * @return Knot object for  knot with specified id.
   */
  public Knot getKnot( int id ) {
    return (Knot)(knots.get( new Integer( id ) ) );
  }



  /**
   * Return Edge object for edge with specified id.
   *
   * @param id ID of the edge.
   * @return Edge object for edge with specified id.
   */
  public Edge getEdge( int id ) {
    return (Edge)(edges.get( new Integer( id ) ) );
  }

  
}
