import java.awt.*;
import java.awt.image.*;
import java.awt.geom.*;
import javax.swing.*;
import java.util.*;
import com.sun.j3d.utils.applet.MainFrame;
import com.sun.j3d.utils.geometry.Cylinder;
import com.sun.j3d.utils.image.TextureLoader;
import com.sun.j3d.utils.geometry.Box;
import com.sun.j3d.utils.behaviors.vp.*;
import com.sun.j3d.utils.universe.*;
import javax.media.j3d.*;
import javax.vecmath.*;

/**
 * Class for visualizing net data. Generates a basic 2d-visualization of a graph (knots and edges), and uses
 * this visualization as texture for a polygon in 3d. Then, additional information is visualized using
 * cylinders based on this polygon and extending in z-direction.
 *
 * @author Stefan Daschek
 */
public class NetVis extends JComponent {

  /** Size (in pixel) of border around 2D-visualization. */
  final static int BORDER = 25;

  /** Size (in pixel) of generated 2d-visualization texture. */
  final static int TEXTURESIZE = 512;

  /** Size (in pixel) of circles representing knots in 2d-visualization. */
  final static int KNOTSIZE = 16;

  /** Size (in pixel) of pie-charts used in 2d-visualization. */
  final static int PIERADIUS = 48;
  
  /** Radius (in world-coordinates) of cylinders used for 3d-visualization. */
  final static float CYLRADIUS = 0.08f;

  /** Maximum height (in world-coordinates) of cylinders used for 3d-visualization. */
  final static float CYLHEIGHT = 0.8f;

  /** Stroke used for visualizing edges. */
  Stroke edgeStroke = new BasicStroke( 5 );

  /** Paint used for visualizing edges. */
  Paint  edgePaint  = Color.black;

  /** Stroke used for visualizing knots. */
  Stroke knotStroke = new BasicStroke( 2 );

  /** Stroke used for visualizing pies. */
  Stroke pieStroke  = new BasicStroke( 2 );

  /** Paint used for clearing pies. */
  Paint  clearPaint = Color.white;

  /** Paint used for outline of pies. */
  Paint  piePaint   = new Color( 128, 128, 0 );

  /** Paint used for fill of pies. */
  Paint  pie2Paint  = new Color( 192, 192, 0 );

  /** Flag whether pies should be drawn. */
  boolean show2dPies     = false;

  /** Flag whether cylinders visualizing load data should be drawn. */
  boolean show3dLoad     = false;

  /** Flag whether cylinders visualizing capacity data should be drawn. */
  boolean show3dCapacity = false;

  /** Data to be visualized. */
  NetData netData;

  
  Canvas3D c3D;
  private SimpleUniverse u = null;
  TransformGroup objTrans;
  float angle = 0.0f;
  Transform3D trans = new Transform3D();
  Transform3D t3d3;
  Texture tex;
  BufferedImage vis2DImg;
  Switch sw;


  /**
   * Default constructor. Initializes netData with <code>null</code>.
   */
  public NetVis() {
    netData = null;
  }



  /**
   * Creates scene graph for Java3D.
   *
   * @return BranchGroup consisting of all elements in the scene graph.
   */
  public BranchGroup createSceneGraph() {
    // Create the root of the branch graph
    BranchGroup objRoot = new BranchGroup();

    // Create a Transformgroup to scale all objects so they
    // appear in the scene.
    TransformGroup objScale = new TransformGroup();
    Transform3D t3d = new Transform3D();
    t3d.setScale( 0.7 );
    objScale.setTransform( t3d );
    objRoot.addChild( objScale );
    
    objTrans = new TransformGroup();
    //write-enable for behaviors
    objTrans.setCapability( TransformGroup.ALLOW_TRANSFORM_WRITE );
    objTrans.setCapability( TransformGroup.ALLOW_TRANSFORM_READ );
    objTrans.setCapability( TransformGroup.ENABLE_PICK_REPORTING );
    t3d3 = new Transform3D();
    t3d3.rotX( -0.8 );
    objTrans.setTransform( t3d3 );
    objScale.addChild( objTrans );

    // bounds
    BoundingSphere bounds = new BoundingSphere( new Point3d( 0.0, 0.0, 0.0 ), 100.0 );

    // set up some light
    Color3f  lColor1 = new Color3f( 0.4f, 0.4f, 0.4f );
    Vector3f lDir1   = new Vector3f( 3.0f, 0.2f, 0.3f );
    Color3f  alColor = new Color3f( 0.6f, 0.6f, 0.6f );
    AmbientLight aLgt = new AmbientLight( alColor );
    aLgt.setInfluencingBounds( bounds );
    DirectionalLight lgt1 = new DirectionalLight( lColor1, lDir1 );
    lgt1.setInfluencingBounds( bounds );
    objRoot.addChild( aLgt );
    objRoot.addChild( lgt1 );

    // Create texture based on 2d-visualization data
    Appearance ap = new Appearance();
    ap.setCapability( Appearance.ALLOW_TEXTURE_WRITE );
    tex = new Texture2D( Texture.BASE_LEVEL, Texture.RGB, TEXTURESIZE, TEXTURESIZE );
    tex.setCapability( Texture.ALLOW_IMAGE_WRITE );
    tex.setBoundaryModeS( Texture.CLAMP );
    tex.setBoundaryModeT( Texture.CLAMP );
    tex.setBoundaryColor( 1.0f, 1.0f, 1.0f, 1.0f );
    tex.setImage( 0, new ImageComponent2D( ImageComponent.FORMAT_RGB, vis2DImg, true, true ) );
    ap.setTexture( tex );
    TextureAttributes texAttr = new TextureAttributes();
    texAttr.setTextureMode( TextureAttributes.MODULATE );
    ap.setTextureAttributes( texAttr );

    // very flat box for holding texture
    Box boxObj = new Box( 1.0f, 1.0f, 0.001f, Box.GENERATE_NORMALS | Box.GENERATE_TEXTURE_COORDS, ap );
    boxObj.setAppearance( ap );
    objTrans.addChild( boxObj );

    // we need a switch to selectively enable/disable 3d-visualizations
    sw = new Switch( Switch.CHILD_NONE );
    sw.setCapability( Switch.ALLOW_SWITCH_WRITE );

    // 3d Load
    Group grp = new Group();
    for( Iterator i = netData.getKnots().values().iterator(); i.hasNext(); ) {
      Knot k = (Knot)i.next();
      TransformGroup cylTrans = new TransformGroup();
      Transform3D t3d2 = new Transform3D();
      t3d2.rotX( Math.PI / 2 );
      t3d2.setTranslation( new Vector3d( (k.x - 50) / 55.0, (k.y - 50) / 55.0, CYLHEIGHT * k.val1 / 200.0f ) );
      cylTrans.setTransform( t3d2 );
      Cylinder cyl = new Cylinder( CYLRADIUS, CYLHEIGHT * k.val1 / 100.0f );
      Appearance ap2 = new Appearance();
      float col = k.val1 / 100.0f;
      ap2.setMaterial( new Material( new Color3f( col, 0.0f, 1 - col ),
                                     new Color3f( Color.black ),
                                     new Color3f( col, 0.0f, 1 - col ),
                                     new Color3f( Color.white ),
                                     32.0f ) );
      cyl.setAppearance( ap2 );
      cylTrans.addChild( cyl );
      grp.addChild( cylTrans );
    }    
    sw.addChild( grp );

    // 3d Capacity
    grp = new Group();
    for( Iterator i = netData.getKnots().values().iterator(); i.hasNext(); ) {
      Knot k = (Knot)i.next();
      TransformGroup cylTrans = new TransformGroup();
      Transform3D t3d2 = new Transform3D();
      t3d2.rotX( Math.PI / 2 );
      t3d2.setTranslation( new Vector3d( (k.x - 50) / 55.0, (k.y - 50) / 55.0, CYLHEIGHT * k.val2 / 200.0f ) );
      cylTrans.setTransform( t3d2 );
      Cylinder cyl = new Cylinder( CYLRADIUS, CYLHEIGHT * k.val2 / 100.0f );
      Appearance ap2 = new Appearance();
      float col = k.val2 / 100.0f;
      ap2.setMaterial( new Material( new Color3f( 0.0f, 0.8f, 0.0f ),
                                     new Color3f( Color.black ),
                                     new Color3f( 0.0f, 0.8f, 0.0f ),
                                     new Color3f( Color.white ),
                                     32.0f ) );
      cyl.setAppearance( ap2 );
      cylTrans.addChild( cyl );
      grp.addChild( cylTrans );
    }    
    sw.addChild( grp );
    
    // 3d Load + Capacity
    grp = new Group();
    TransparencyAttributes tr = new TransparencyAttributes();
    tr.setTransparency( 0.5f );
    tr.setTransparencyMode( tr.NICEST );
    for( Iterator i = netData.getKnots().values().iterator(); i.hasNext(); ) {
      Knot k = (Knot)i.next();
      float height1 = CYLHEIGHT * k.val2 / 200.0f;
      float height2 = height1 * k.val1 / 100.0f;
      
      TransformGroup cylTrans = new TransformGroup();
      Transform3D t3d2 = new Transform3D();
      t3d2.rotX( Math.PI / 2 );
      t3d2.setTranslation( new Vector3d( (k.x - 50) / 55.0, (k.y - 50) / 55.0, height1 + height2 ) );
      cylTrans.setTransform( t3d2 );
      Cylinder cyl = new Cylinder( CYLRADIUS, (height1 - height2) * 2.0f );
      Appearance ap2 = new Appearance();
      float col = k.val2 / 100.0f;
      ap2.setMaterial( new Material( new Color3f( 0.7f, 0.7f, 0.7f ),
                                     new Color3f( Color.black ),
                                     new Color3f( 0.7f, 0.7f, 0.7f ),
                                     new Color3f( Color.white ),
                                     16.0f ) );
      ap2.setTransparencyAttributes( tr );
      cyl.setAppearance( ap2 );
      cylTrans.addChild( cyl );
      grp.addChild( cylTrans );
      cylTrans = new TransformGroup();
      t3d2 = new Transform3D();
      t3d2.rotX( Math.PI / 2 );
      t3d2.setTranslation( new Vector3d( (k.x - 50) / 55.0, (k.y - 50) / 55.0, height2 ) );
      cylTrans.setTransform( t3d2 );
      cyl = new Cylinder( CYLRADIUS, height2 * 2.0f );
      ap2 = new Appearance();
      col = k.val1 / 100.0f;
      ap2.setMaterial( new Material( new Color3f( col, 0.0f, 1 - col ),
                                     new Color3f( Color.black ),
                                     new Color3f( col, 0.0f, 1 - col ),
                                     new Color3f( Color.white ),
                                     32.0f ) );
      cyl.setAppearance( ap2 );
      cylTrans.addChild( cyl );
      grp.addChild( cylTrans );
    }    
    sw.addChild( grp );
    
    objTrans.addChild( sw );
    
    return objRoot;
  }



  /**
   * Does basic setup of Java3D, create 2d visualization and 3d scene graph.
   */
  public void setup() {

    // Create buffer for texture holding 2d visualization and initialize it.
    vis2DImg = (BufferedImage)createImage( TEXTURESIZE, TEXTURESIZE );
    visualize2D( vis2DImg );

    // Create Canvas3D and add it to the UI.
    c3D = new Canvas3D( SimpleUniverse.getPreferredConfiguration() );
    setLayout( new BorderLayout() );
    add( "Center", c3D );

    // Create scene graph.
    BranchGroup scene = createSceneGraph();
    scene.setCapability( BranchGroup.ALLOW_BOUNDS_READ );
    u = new SimpleUniverse(c3D);

    // This will move the ViewPlatform back a bit so the
    // objects in the scene can be viewed.
    u.getViewingPlatform().setNominalViewingTransform();

    // add orbit behavior but disable translate
    OrbitBehavior orbit = new OrbitBehavior( c3D, OrbitBehavior.REVERSE_ALL );
    BoundingSphere bounds = new BoundingSphere(new Point3d(0.0, 0.0, 0.0), 100.0);
    orbit.setSchedulingBounds(bounds);
    orbit.setRotXFactor( 1.0 );
    orbit.setRotYFactor( 0.5 );
    u.getViewingPlatform().setViewPlatformBehavior(orbit);

    u.addBranchGraph(scene);
  }



  /**
   * Set data to be visualized and update scene graph and 2d visualization.
   *
   * @param netData NetData object to be visualized.
   */
  public void setNetData( NetData netData ) {
    this.netData = netData;
    setup();
  }



  /**
   * Do 2d visualization of graph. Visualize knots, edges and piecharts.
   *
   * @param img BufferedImg used for constructing the visualization.
   */
  protected void visualize2D( BufferedImage img ) {
    if( (null == netData) || (null == img) )
      return;

    Graphics2D g2  = (Graphics2D)img.getGraphics();
    int width      = img.getWidth() - BORDER * 2;
    int height     = img.getHeight() - BORDER * 2;

    g2.setPaint( clearPaint );
    g2.fillRect( 0, 0, width + BORDER * 2, height +  BORDER * 2 );
    g2.setColor( Color.black );
    
    g2.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON );

    // Draw edges
    g2.setStroke( edgeStroke );
    g2.setPaint( edgePaint );
    for( Iterator i = netData.getEdges().values().iterator(); i.hasNext(); ) {
      Edge e = (Edge)i.next();
      double x1 = netData.getKnot( e.from ).x / 100.0 * width + BORDER;
      double y1 = netData.getKnot( e.from ).y / 100.0 * height + BORDER;
      double x2 = netData.getKnot( e.to   ).x / 100.0 * width + BORDER;
      double y2 = netData.getKnot( e.to   ).y / 100.0 * height + BORDER;
      g2.draw( new Line2D.Double( x1, y1, x2, y2 ) );

    }    

    // Draw knots
    g2.setStroke( knotStroke );
    for( Iterator i = netData.getKnots().values().iterator(); i.hasNext(); ) {
      Knot k = (Knot)i.next();
      Ellipse2D knot = new Ellipse2D.Double( k.x / 100.0 * width - KNOTSIZE / 2 + BORDER,
                                             k.y / 100.0 * height - KNOTSIZE / 2 + BORDER,
                                             KNOTSIZE, KNOTSIZE );
      g2.setColor( Color.white );
      g2.fill( knot );
      g2.setColor( Color.black );
      g2.draw( knot );
    }

    // Draw pies
    g2.setStroke( pieStroke );
    if( show2dPies ) {
      for( Iterator i = netData.getEdges().values().iterator(); i.hasNext(); ) {
        Edge e = (Edge)i.next();
        double x1 = netData.getKnot( e.from ).x / 100.0 * width + BORDER;
        double y1 = netData.getKnot( e.from ).y / 100.0 * height + BORDER;
        double x2 = netData.getKnot( e.to   ).x / 100.0 * width + BORDER;
        double y2 = netData.getKnot( e.to   ).y / 100.0 * height + BORDER;
        double x3 = (x1 + x2 - PIERADIUS) / 2.0;
        double y3 = (y1 + y2 - PIERADIUS) / 2.0;
        Ellipse2D pie = new Ellipse2D.Double( x3, y3, PIERADIUS, PIERADIUS );
        g2.setPaint( clearPaint );
        g2.fill( pie );
        g2.setPaint( new Color( e.val / 100.0f, 0.7f, 0.0f ) );
        g2.fill( new Arc2D.Double( x3, y3, PIERADIUS, PIERADIUS, 0, 3.6 * e.val, Arc2D.PIE ) );
        g2.setPaint( piePaint );
        g2.draw( pie );
      }
    }
  }



  /**
   * Toggle visualization of load data.
   *
   * @param show True if load data should be visualized.
   */
  public void setShow3dLoad( boolean show ) {
    show3dLoad = show;
    updateShow3d();
  }



  /**
   * Toggle visualization of capacity data.
   *
   * @param show True if capacity data should be visualized.
   */
  public void setShow3dCapacity( boolean show ) {
    show3dCapacity = show;
    updateShow3d();
  }



  /**
   * Update scene graph to show correct objects according to flags <code>show3dLoad</code> and
   * <code>show3dCapacity</code>.
   */
  protected void updateShow3d() {
    if( show3dLoad && show3dCapacity )
      sw.setWhichChild( 2 );
    else if( show3dLoad )
      sw.setWhichChild( 0 );
    else if( show3dCapacity )
      sw.setWhichChild( 1 );
    else
      sw.setWhichChild( sw.CHILD_NONE );
  }


  /**
   * Toggle visualization of 2d pies.
   *
   * @param show True if 2d pies should be visualized.
   */
  public void setShow2dPies( boolean show ) {
    show2dPies = show;
    visualize2D( vis2DImg );
    if( tex != null )
      tex.setImage( 0, new ImageComponent2D( ImageComponent.FORMAT_RGB, vis2DImg, true, true ) );
  }



  /**
   * Reset viewing transform of scene graph to initial value.
   */
  public void reset3d() {
    u.getViewingPlatform().setNominalViewingTransform();
  }
}
